#include "mamalleg.h"
#include "driver.h"
#include <sys/farptr.h>
#include <go32.h>
#include "dirty.h"
/* for VGA triple buffer - we need the register structure */
#include "gen15khz.h"

/* from video.c (required for 15.75KHz Arcade Monitor Modes) */
extern int half_yres;
extern int unchained;

extern char *dirty_old;
extern char *dirty_new;

extern struct osd_bitmap *scrbitmap;
extern int gfx_xoffset;
extern int gfx_yoffset;
extern int gfx_display_lines;
extern int gfx_display_columns;
extern int gfx_width;
extern int gfx_height;
extern int skiplines;
extern int skipcolumns;
extern int use_triplebuf;
extern int triplebuf_pos,triplebuf_page_width;

unsigned int doublepixel[256];
unsigned int quadpixel[256]; /* for quadring pixels */

// Start Emu Plus
unsigned char *redpoin;
unsigned char *grnpoin;
unsigned char *blupoin;
int flippy = 0;
int runnit;
extern int blur_yes;
// End Emu Plus


/* current 'page' for unchained modes */
static int xpage = -1;
/*default page sizes for non-triple buffering */
int xpage_size = 0x8000;
int no_xpages = 2;

/* this function lifted from Allegro */
static int vesa_scroll_async(int x, int y)
{
   int ret;
	#define BYTES_PER_PIXEL(bpp)     (((int)(bpp) + 7) / 8)	/* in Allegro */
	extern __dpmi_regs _dpmi_reg;	/* in Allegro... I think */

//   vesa_xscroll = x;
//   vesa_yscroll = y;

#if 0
   int seg;
   long a;
	extern void (*_pm_vesa_scroller)(void);	/* in Allegro */
	extern int _mmio_segment;	/* in Allegro */
   if (_pm_vesa_scroller) {            /* use protected mode interface? */
      seg = _mmio_segment ? _mmio_segment : _my_ds();

      a = ((x * BYTES_PER_PIXEL(screen->vtable->color_depth)) +
	   (y * ((unsigned long)screen->line[1] - (unsigned long)screen->line[0]))) / 4;

      asm (
	 "  pushw %%es ; "
	 "  movw %w1, %%es ; "         /* set the IO segment */
	 "  call *%0 ; "               /* call the VESA function */
	 "  popw %%es "

      :                                /* no outputs */

      : "S" (_pm_vesa_scroller),       /* function pointer in esi */
	"a" (seg),                     /* IO segment in eax */
	"b" (0x00),                    /* mode in ebx */
	"c" (a & 0xFFFF),              /* low word of address in ecx */
	"d" (a >> 16)                  /* high word of address in edx */

//      : "memory", "%edi", "%cc"        /* clobbers edi and flags */
		: "memory", "%ebp", "%edi", "%cc" /* clobbers ebp, edi and flags (at least) */
      );

      ret = 0;
   }
   else
#endif
   {                              /* use a real mode interrupt call */
      _dpmi_reg.x.ax = 0x4F07;
      _dpmi_reg.x.bx = 0;
      _dpmi_reg.x.cx = x;
      _dpmi_reg.x.dx = y;

      __dpmi_int(0x10, &_dpmi_reg);
      ret = _dpmi_reg.h.ah;

//      _vsync_in();
   }

   return (ret ? -1 : 0);
}


void blitscreen_dirty1_vga(void)
{
	int width4, x, y;
	unsigned long *lb, address;

	width4 = (scrbitmap->line[1] - scrbitmap->line[0]) / 4;
	address = 0xa0000 + gfx_xoffset + gfx_yoffset * gfx_width;
	lb = (unsigned long *)(scrbitmap->line[skiplines] + skipcolumns);

	for (y = 0; y < gfx_display_lines; y += 16)
	{
		for (x = 0; x < gfx_display_columns; )
		{
			int w = 16;
			if (ISDIRTY(x,y))
			{
				unsigned long *lb0 = lb + x / 4;
				unsigned long address0 = address + x;
				int h;
				while (x + w < gfx_display_columns && ISDIRTY(x+w,y))
                    w += 16;
				if (x + w > gfx_display_columns)
                    w = gfx_display_columns - x;
				for (h = 0; h < 16 && y + h < gfx_display_lines; h++)
				{
					_dosmemputl(lb0, w/4, address0);
					lb0 += width4;
					address0 += gfx_width;
				}
			}
			x += w;
        }
		lb += 16 * width4;
		address += 16 * gfx_width;
	}
}

void blitscreen_dirty0_vga(void)
{
	int width4,y,columns4;
	unsigned long *lb, address;

	width4 = (scrbitmap->line[1] - scrbitmap->line[0]) / 4;
	columns4 = gfx_display_columns/4;
	address = 0xa0000 + gfx_xoffset + gfx_yoffset * gfx_width;
	lb = (unsigned long *)(scrbitmap->line[skiplines] + skipcolumns);
	for (y = 0; y < gfx_display_lines; y++)
	{
		_dosmemputl(lb,columns4,address);
		lb+=width4;
		address+=gfx_width;
	}
}

/* for flipping between the unchained VGA pages */
INLINE void unchained_flip(void)
{
	int	flip_value1,flip_value2;
	int	temp;

/* memory address of non-visible page */
	temp = xpage_size * xpage;

	flip_value1 = ((temp & 0xff00) | 0x0c);
	flip_value2 = (((temp << 8) & 0xff00) | 0x0d);

/* flip the page var */
	xpage ++;
	if (xpage == no_xpages)
		xpage = 0;
/* need to change the offset address during the active display interval */
/* as the value is read during the vertical retrace */
	__asm__ __volatile__ (

	"movw   $0x3da,%%dx \n"
	"cli \n"
	".align 4                \n"
/* check for active display interval */
	"0:\n"
	"inb    %%dx,%%al \n"
	"testb  $1,%%al \n"
	"jz  	0b \n"
/* change the offset address */
	"movw   $0x3d4,%%dx \n"
	"movw   %%cx,%%ax \n"
	"outw   %%ax,%%dx \n"
	"movw   %%bx,%%ax \n"
	"outw   %%ax,%%dx \n"
	"sti \n"


/* outputs  (none)*/
	:"=c" (flip_value1),
	"=b" (flip_value2)
/* inputs -
 ecx = flip_value1 , ebx = flip_value2 */
	:"0" (flip_value1),
	"1" (flip_value2)
/* registers modified */
	:"ax", "dx", "cc", "memory"
	);
}



/* unchained dirty modes */
void blitscreen_dirty1_unchained_vga(void)
{
	int x, y, i, outval, dirty_page, triple_page, write_triple;
	int plane, planeval, iloop, page;
	unsigned long *lb, address, triple_address;
	unsigned char *lbsave;
	unsigned long asave, triple_save;
	static int width4, word_blit, dirty_height;
	static int source_width, source_line_width, dest_width, dest_line_width;

	/* calculate our statics on the first call */
	if (xpage == -1)
	{
		width4 = ((scrbitmap->line[1] - scrbitmap->line[0]) >> 2);
		source_width = width4 << half_yres;
		dest_width = gfx_width >> 2;
		dest_line_width = (gfx_width << 2) >> half_yres;
		source_line_width = width4 << 4;
		xpage = 1;
		dirty_height = 16 >> half_yres;
		/* check if we have to do word rather than double word updates */
		word_blit = ((gfx_display_columns >> 2) & 3);
	}

	/* setup or selector */
	_farsetsel(screen->seg);

	dirty_page = xpage;
	/* non visible page, if we're triple buffering */
	triple_page = xpage + 1;
	if (triple_page == no_xpages)
		triple_page = 0;

	/* need to update all 'pages', but only update each page while it isn't visible */
	for (page = 0; page < 2; page ++)
	{
		planeval=0x0100;
		write_triple = (!page | (no_xpages == 3));

		/* go through each bit plane */
		for (plane = 0; plane < 4 ;plane ++)
		{
			address = 0xa0000 + (xpage_size * dirty_page)+(gfx_xoffset >> 2) + (((gfx_yoffset >> half_yres) * gfx_width) >> 2);
			triple_address = 0xa0000 + (xpage_size * triple_page)+(gfx_xoffset >> 2) + (((gfx_yoffset >> half_yres) * gfx_width) >> 2);
			lb = (unsigned long *)(scrbitmap->line[skiplines] + skipcolumns + plane);
			/*set the bit plane */
			outportw(0x3c4, planeval|0x02);
			for (y = 0; y < gfx_display_lines ; y += 16)
			{
				for (x = 0; x < gfx_display_columns; )
				{
					int w = 16;
					if (ISDIRTY(x,y))
					{
						unsigned long *lb0 = lb + (x >> 2);
						unsigned long address0 = address + (x >> 2);
						unsigned long address1 = triple_address + (x >> 2);
						int h;
						while (x + w < gfx_display_columns && ISDIRTY(x+w,y))
							w += 16;
						if (x + w > gfx_display_columns)
							w = gfx_display_columns - x;
						if(word_blit)
						{
							iloop = w >> 3;
							for (h = 0; h < dirty_height && (y + (h << half_yres)) < gfx_display_lines; h++)
							{
								asave = address0;
								triple_save = address1;
								lbsave = (unsigned char *)lb0;
								for(i = 0; i < iloop; i++)
								{
									outval = *lbsave | (lbsave[4] << 8);
									_farnspokew(asave, outval);
									/* write 2 pages on first pass if triple buffering */
									if (write_triple)
									{
										_farnspokew(triple_save, outval);
										triple_save += 2;
									}
									lbsave += 8;
									asave += 2;
								}
								if (w&4)
								{
									_farnspokeb(asave, *lbsave);
		   							if (write_triple)
										_farnspokeb(triple_save, *lbsave);
								}


								lb0 += source_width;
								address0 += dest_width;
								address1 += dest_width;
							}
						}
						else
						{
							iloop = w >> 4;
							for (h = 0; h < dirty_height && (y + (h << half_yres)) < gfx_display_lines; h++)
							{
								asave = address0;
								triple_save = address1;
								lbsave = (unsigned char *)lb0;
								for(i = 0; i < iloop; i++)
								{
									outval = *lbsave | (lbsave[4] << 8) | (lbsave[8] << 16) | (lbsave[12] << 24);
									_farnspokel(asave, outval);
									/* write 2 pages on first pass if triple buffering */
									if (page == 0 && no_xpages == 3)
									{
										_farnspokel(triple_save, outval);
										triple_save += 4;
									}
									lbsave += 16;
									asave += 4;
								}
								lb0 += source_width;
								address0 += dest_width;
								address1 += dest_width;
							}
						}
					}
					x += w;
				}
				lb += source_line_width;
				address += dest_line_width;
				triple_address += dest_line_width;
			}
			/* move onto the next bit plane */
			planeval <<= 1;
		}
		/* 'bank switch' our unchained output on the first loop */
		if(!page)
			unchained_flip();
		/* move onto next 'page' */
		dirty_page += (no_xpages - 1);
		if (dirty_page >= no_xpages)
			dirty_page -= no_xpages;
	}
}

void init_unchained_blit(void)
{
    xpage=-1;
}

/* Macros for non dirty unchained blits */
#define UNCHAIN_BLIT_START \
	int dummy1,dummy2,dummy3; \
	__asm__ __volatile__ ( \
/* save es and set it to our video selector */ \
	"pushw  %%es \n" \
	"movw   %%dx,%%es \n" \
	"movw   $0x102,%%ax \n" \
	"cld \n" \
	".align 4 \n" \
/* --bit plane loop-- */ \
	"0:\n" \
/* save everything */ \
	"pushl  %%ebx \n" \
	"pushl  %%edi \n" \
	"pushl  %%eax \n" \
	"pushl  %%ecx \n" \
/* set the bit plane */ \
	"movw   $0x3c4,%%dx \n" \
	"outw   %%ax,%%dx \n" \
/* edx now free, so use it for the memwidth */ \
	"movl   %4,%%edx \n" \
/* --height loop-- */ \
	"1:\n" \
/* save counter , source + dest address */ \
	"pushl 	%%ecx \n" \
	"pushl	%%ebx \n" \
	"pushl	%%edi \n" \
/* --width loop-- */ \
	"movl   %3,%%ecx \n" \
	"2:\n" \
/* get the 4 bytes */ \
/* bswap should be faster than shift, */ \
/* movl should be faster then movb */ \
	"movl   %%ds:12(%%ebx),%%eax \n" \
	"movb   %%ds:8(%%ebx),%%ah \n" \
	"bswap  %%eax \n" \
	"movb   %%ds:(%%ebx),%%al \n" \
	"movb   %%ds:4(%%ebx),%%ah \n" \
/* write the thing to video */ \
	"stosl \n" \
/* move onto next source address */ \
	"addl   $16,%%ebx \n" \
	"loop	2b \n"

#define UNCHAIN_BLIT_END \
/* --end of width loop-- */ \
/* get counter, source + dest address back */ \
	"popl	%%edi \n" \
	"popl	%%ebx \n" \
	"popl   %%ecx \n" \
/* move to the next line */ \
	"addl   %%edx,%%ebx \n" \
	"addl   %%esi,%%edi \n" \
	"loop   1b \n" \
/* --end of height loop-- */ \
/* get everything back */ \
	"popl   %%ecx \n" \
	"popl   %%eax \n"  \
	"popl   %%edi \n" \
	"popl   %%ebx \n" \
/* move onto next bit plane */ \
	"incl   %%ebx \n" \
	"shlb   $1,%%ah \n" \
/* check if we've done all 4 or not */ \
	"testb  $0x10,%%ah \n" \
	"jz		0b \n" \
/* --end of bit plane loop-- */ \
/* restore es */ \
	"popw   %%es \n" \
/* outputs (aka clobbered input) */ \
	:"=S" (dummy1), \
	"=b" (dummy2), \
	"=d" (dummy3) \
/* inputs */ \
/* %0=width, %1=memwidth, */ \
/* esi = scrwidth */ \
/* ebx = src, ecx = height */ \
/* edx = seg, edi = address */ \
	:"g" (width), \
	"g" (memwidth), \
	"c" (height), \
	"D" (address), \
	"0" (scrwidth), \
	"1" (src), \
	"2" (seg) \
/* registers modified */ \
	:"ax", "cc", "memory" \
	);



/*asm routine for unchained blit - writes 4 bytes at a time */
INLINE void unchain_dword_blit(unsigned long *src,short seg,unsigned long address,int width,int height,int memwidth,int scrwidth)
{
/* straight forward double word blit */
	UNCHAIN_BLIT_START
	UNCHAIN_BLIT_END
}

/*asm routine for unchained blit - writes 4 bytes at a time, then 2 'odd' bytes at the end of each row */
INLINE void unchain_word_blit(unsigned long *src,short seg,unsigned long address,int width,int height,int memwidth,int scrwidth)
{
	UNCHAIN_BLIT_START
/* get the extra 2 bytes at end of the row */
	"movl   %%ds:(%%ebx),%%eax \n"
	"movb   %%ds:4(%%ebx),%%ah \n"
/* write the thing to video */
	"stosw \n"
	UNCHAIN_BLIT_END
}

/*asm routine for unchained blit - writes 4 bytes at a time, then 1 'odd' byte at the end of each row */
INLINE void unchain_byte_blit(unsigned long *src,short seg,unsigned long address,int width,int height,int memwidth,int scrwidth)
{
	UNCHAIN_BLIT_START
/* get the extra byte at end of the row */
	"movb   %%ds:(%%ebx),%%eax \n"
/* write the thing to video */
	"stosb \n"
	UNCHAIN_BLIT_END
}

/* unchained 'non-dirty' modes */
void blitscreen_dirty0_unchained_vga(void)
{
	unsigned long *lb, address;
	static int width4,columns4,column_chained,memwidth,scrwidth,disp_height,blit_type;

   /* only calculate our statics the first time around */
	if(xpage==-1)
	{
		width4 = (scrbitmap->line[1] - scrbitmap->line[0]) >> 2;
		columns4 = gfx_display_columns >> 2;
		disp_height = gfx_display_lines >> half_yres;

		xpage = 1;
		memwidth = (scrbitmap->line[1] - scrbitmap->line[0]) << half_yres;
		scrwidth = gfx_width >> 2;

		/* check for 'not divisible by 8' */
		if((columns4 & 1))
			blit_type = 2;    /* byte blit */
		else
		{
			if((columns4 & 2))
				blit_type = 1; /* word blit */
			else
				blit_type = 0; /* double word blit */
		}

		column_chained = columns4 >> 2;
	}

	/* get the start of the screen bitmap */
	lb = (unsigned long *)(scrbitmap->line[skiplines] + skipcolumns);
	/* and the start address in video memory */
	address = 0xa0000 + (xpage_size * xpage)+(gfx_xoffset >> 2) + (((gfx_yoffset >> half_yres) * gfx_width) >> 2);
	/* call the appropriate blit routine */
	switch (blit_type)
	{
		case 0: /* double word blit */
			unchain_dword_blit (lb, screen->seg, address, column_chained, disp_height, memwidth, scrwidth);
			break;
		case 1: /* word blit */
			unchain_word_blit (lb, screen->seg, address, column_chained, disp_height, memwidth, scrwidth);
			break;
		case 2: /* byte blit */
			unchain_byte_blit (lb, screen->seg, address, column_chained, disp_height, memwidth, scrwidth);
			break;
	}
	/* 'bank switch' our unchained output */
	unchained_flip();
}



/* setup register array to be unchained */
void unchain_vga(Register *pReg)
{
/* setup registers for an unchained mode */
	pReg[UNDERLINE_LOC_INDEX].value = 0x00;
	pReg[MODE_CONTROL_INDEX].value = 0xe3;
	pReg[MEMORY_MODE_INDEX].value = 0x06;
/* flag the fact it's unchained */
	unchained = 1;
}


INLINE void copyline_1x_8bpp(unsigned char *src,short seg,unsigned long address,int width4)
{
	short src_seg;

	src_seg = _my_ds();

	_movedatal(src_seg,(unsigned long)src,seg,address,width4);
}

INLINE void copyline_1x_16bpp(unsigned char *src,short seg,unsigned long address,int width2)
{
	short src_seg;

	src_seg = _my_ds();

	_movedatal(src_seg,(unsigned long)src,seg,address,width2);
}

#if 1 /* use the C approach instead */
INLINE void copyline_2x_8bpp(unsigned char *src,short seg,unsigned long address,int width4)
{
	__asm__ __volatile__ (
	"pushw %%es              \n"
	"movw %%dx, %%es         \n"
	"cld                     \n"
	".align 4                \n"
	"0:                      \n"
	"lodsl                   \n"
	"movl %%eax, %%ebx       \n"
	"bswap %%eax             \n"
	"xchgw %%ax,%%bx         \n"
	"roll $8, %%eax          \n"
	"stosl                   \n"
	"movl %%ebx, %%eax       \n"
	"rorl $8, %%eax          \n"
	"stosl                   \n"
	"loop 0b                 \n"
	"popw %%ax               \n"
	"movw %%ax, %%es         \n"
	:
	"=c" (width4),
	"=d" (seg),
	"=S" (src),
	"=D" (address)
	:
	"0" (width4),
	"1" (seg),
	"2" (src),
	"3" (address):
	"ax", "bx", "cc", "memory");
}
#else
INLINE void copyline_2x_8bpp(unsigned char *src,short seg,unsigned long address,int width4)
{
	int i;

	/* set up selector */
	_farsetsel(seg);

	for (i = width; i > 0; i--)
	{
		_farnspokel (address, doublepixel[*src] | (doublepixel[*(src+1)] << 16));
		_farnspokel (address+4, doublepixel[*(src+2)] | (doublepixel[*(src+3)] << 16));
		address+=8;
		src+=4;
	}
}
#endif

INLINE void copyline_2x_16bpp(unsigned char *src,short seg,unsigned long address,int width2)
{
	__asm__ __volatile__ (
	"pushw %%es              \n"
	"movw %%dx, %%es         \n"
	"cld                     \n"
	".align 4                \n"
	"0:                      \n"
	"lodsl                   \n"
	"movl %%eax, %%ebx       \n"
	"roll $16, %%eax         \n"
	"xchgw %%ax,%%bx         \n"
	"stosl                   \n"
	"movl %%ebx, %%eax       \n"
	"stosl                   \n"
	"loop 0b                 \n"
	"popw %%ax               \n"
	"movw %%ax, %%es         \n"
	:
	"=c" (width2),
	"=d" (seg),
	"=S" (src),
	"=D" (address)
	:
	"0" (width2),
	"1" (seg),
	"2" (src),
	"3" (address):
	"ax", "bx", "cc", "memory");
}

INLINE void copyline_3x_8bpp(unsigned char *src,short seg,unsigned long address,int width4)
{
	int i;

	/* set up selector */
	_farsetsel(seg);

	for (i = width4; i > 0; i--)
	{
		_farnspokel (address, (quadpixel[*src] & 0x00ffffff) | (quadpixel[*(src+1)] & 0xff000000));
		_farnspokel (address+4, (quadpixel[*(src+1)] & 0x0000ffff) | (quadpixel[*(src+2)] & 0xffff0000));
		_farnspokel (address+8, (quadpixel[*(src+2)] & 0x000000ff) | (quadpixel[*(src+3)] & 0xffffff00));
		address+=3*4;
		src+=4;
	}
}

INLINE void copyline_4x_8bpp(unsigned char *src,short seg,unsigned long address,int width4)
{
	int i;

	/* set up selector */
	_farsetsel(seg);

	for (i = width4; i > 0; i--)
	{
		_farnspokel (address, quadpixel[*src]);
		_farnspokel (address+4, quadpixel[*(src+1)]);
		_farnspokel (address+8, quadpixel[*(src+2)]);
		_farnspokel (address+12, quadpixel[*(src+3)]);
		address+=16;
		src+=4;
	}
}



#define DIRTY1(BPP) \
	short dest_seg; \
	int x,y,vesa_line,line_offs,xoffs; \
	unsigned char *lb; \
	unsigned long address; \
	dest_seg = screen->seg; \
	vesa_line = gfx_yoffset; \
	line_offs = scrbitmap->line[1] - scrbitmap->line[0]; \
	xoffs = (BPP/8)*gfx_xoffset; \
	lb = scrbitmap->line[skiplines] + (BPP/8)*skipcolumns; \

#define DIRTY1_NXNS(MX,MY,BPP) \
	for (y = 0;y < gfx_display_lines;y += 16) \
	{ \
		for (x = 0;x < gfx_display_columns; /* */) \
		{ \
			int w = 16; \
			if (ISDIRTY(x,y)) \
            { \
				unsigned char *src = lb + (BPP/8)*x; \
                int vesa_line0 = vesa_line, h; \
				while (x + w < gfx_display_columns && ISDIRTY(x+w,y)) \
                    w += 16; \
				if (x + w > gfx_display_columns) \
					w = gfx_display_columns - x; \
				for (h = 0; h < 16 && y + h < gfx_display_lines; h++) \
                { \
					address = bmp_write_line(screen,vesa_line0) + xoffs + MX*(BPP/8)*x; \
					copyline_##MX##x_##BPP##bpp(src,dest_seg,address,w/(4/(BPP/8))); \
					vesa_line0 += MY; \
					src += line_offs; \
				} \
			} \
			x += w; \
		} \
		vesa_line += MY*16; \
		lb += 16 * line_offs; \
	}

#define DIRTY1_NX2(MX,BPP) \
	for (y = 0;y < gfx_display_lines;y += 16) \
	{ \
		for (x = 0;x < gfx_display_columns; /* */) \
		{ \
			int w = 16; \
			if (ISDIRTY(x,y)) \
            { \
				unsigned char *src = lb + (BPP/8)*x; \
                int vesa_line0 = vesa_line, h; \
				while (x + w < gfx_display_columns && ISDIRTY(x+w,y)) \
                    w += 16; \
				if (x + w > gfx_display_columns) \
					w = gfx_display_columns - x; \
				for (h = 0; h < 16 && y + h < gfx_display_lines; h++) \
                { \
					address = bmp_write_line(screen,vesa_line0) + xoffs + MX*(BPP/8)*x; \
					copyline_##MX##x_##BPP##bpp(src,dest_seg,address,w/(4/(BPP/8))); \
					address = bmp_write_line(screen,vesa_line0+1) + xoffs + MX*(BPP/8)*x; \
					copyline_##MX##x_##BPP##bpp(src,dest_seg,address,w/(4/(BPP/8))); \
					vesa_line0 += 2; \
					src += line_offs; \
				} \
			} \
			x += w; \
		} \
		vesa_line += 16 * 2; \
		lb += 16 * line_offs; \
	}

#define DIRTY1_NX3(MX,BPP) \
	for (y = 0;y < gfx_display_lines;y += 16) \
	{ \
		for (x = 0;x < gfx_display_columns; /* */) \
		{ \
			int w = 16; \
			if (ISDIRTY(x,y)) \
            { \
				unsigned char *src = lb + (BPP/8)*x; \
                int vesa_line0 = vesa_line, h; \
				while (x + w < gfx_display_columns && ISDIRTY(x+w,y)) \
                    w += 16; \
				if (x + w > gfx_display_columns) \
					w = gfx_display_columns - x; \
				for (h = 0; h < 16 && y + h < gfx_display_lines; h++) \
                { \
					address = bmp_write_line(screen,vesa_line0) + xoffs + MX*(BPP/8)*x; \
					copyline_##MX##x_##BPP##bpp(src,dest_seg,address,w/(4/(BPP/8))); \
					address = bmp_write_line(screen,vesa_line0+1) + xoffs + MX*(BPP/8)*x; \
					copyline_##MX##x_##BPP##bpp(src,dest_seg,address,w/(4/(BPP/8))); \
					address = bmp_write_line(screen,vesa_line0+2) + xoffs + MX*(BPP/8)*x; \
					copyline_##MX##x_##BPP##bpp(src,dest_seg,address,w/(4/(BPP/8))); \
					vesa_line0 += 3; \
					src += line_offs; \
				} \
			} \
			x += w; \
		} \
		vesa_line += 16 * 3; \
		lb += 16 * line_offs; \
	}


#define DIRTY0(BPP) \
	short dest_seg; \
	int y,vesa_line,line_offs,xoffs,width; \
	unsigned char *src; \
	unsigned long address; \
	dest_seg = screen->seg; \
	vesa_line = gfx_yoffset; \
	line_offs = (scrbitmap->line[1] - scrbitmap->line[0]); \
	xoffs = (BPP/8)*(gfx_xoffset + triplebuf_pos); \
	width = gfx_display_columns/(4/(BPP/8)); \
	src = scrbitmap->line[skiplines] + (BPP/8)*skipcolumns;	\

#define TRIPLEBUF_FLIP \
	if (use_triplebuf) \
	{ \
		vesa_scroll_async(triplebuf_pos,0); \
		triplebuf_pos = (triplebuf_pos + triplebuf_page_width) % (3*triplebuf_page_width); \
	}

// Start Emu Plus
#define DIRTY0_NXNS(MX,MY,BPP) \
	if (blur_yes) \
	{ \
		runnit = 0;\
		if (flippy == 1)\
		{ \
			src += line_offs;\
			vesa_line += MY; \
			for (y = 0;y < gfx_display_lines/2;y++) \
			{ \
				address = bmp_write_line(screen,vesa_line) + xoffs; \
				copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
				vesa_line += MY*2; \
				src += line_offs*2; \
			} \
			flippy=0 ; runnit = 1; \
		} \
		if (flippy == 0 && runnit == 0)\
		{ \
			for (y = 0;y < gfx_display_lines/2;y++) \
			{ \
				address = bmp_write_line(screen,vesa_line) + xoffs; \
				copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
				vesa_line += MY*2; \
				src += line_offs*2; \
			} \
			flippy=1;\
		} \
	} else { \
		for (y = 0;y < gfx_display_lines;y++) \
		{ \
			address = bmp_write_line(screen,vesa_line) + xoffs; \
			copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
			vesa_line += MY; \
			src += line_offs; \
		} \
	} \
	TRIPLEBUF_FLIP

// End Emu Plus

#define DIRTY0_NX2(MX,BPP) \
	for (y = 0;y < gfx_display_lines;y++) \
	{ \
		address = bmp_write_line(screen,vesa_line) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
		address = bmp_write_line(screen,vesa_line+1) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
		vesa_line += 2; \
		src += line_offs; \
	} \
	TRIPLEBUF_FLIP

#define DIRTY0_NX3(MX,BPP) \
	for (y = 0;y < gfx_display_lines;y++) \
	{ \
		address = bmp_write_line(screen,vesa_line) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
		address = bmp_write_line(screen,vesa_line+1) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
		address = bmp_write_line(screen,vesa_line+2) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
		vesa_line += 3; \
		src += line_offs; \
	} \
	TRIPLEBUF_FLIP



void blitscreen_dirty1_vesa_1x_1x_8bpp(void)   { DIRTY1(8)  DIRTY1_NXNS(1,1,8)  }
void blitscreen_dirty1_vesa_1x_2x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX2 (1,8)  }
void blitscreen_dirty1_vesa_1x_2xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(1,2,8)  }
void blitscreen_dirty1_vesa_2x_1x_8bpp(void)   { DIRTY0(8)  DIRTY0_NXNS(2,1,8)  }
void blitscreen_dirty1_vesa_2x_2x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX2 (2,8)  }
void blitscreen_dirty1_vesa_2x_2xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(2,2,8)  }
void blitscreen_dirty1_vesa_2x_3x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX3 (2,8)  }
void blitscreen_dirty1_vesa_2x_3xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(2,3,8)  }
void blitscreen_dirty1_vesa_3x_2x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX2 (3,8)  }
void blitscreen_dirty1_vesa_3x_2xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(3,2,8)  }
void blitscreen_dirty1_vesa_3x_3x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX3 (3,8)  }
void blitscreen_dirty1_vesa_3x_3xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(3,3,8)  }
void blitscreen_dirty1_vesa_4x_2x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX2 (4,8)  }
void blitscreen_dirty1_vesa_4x_2xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(4,2,8)  }
void blitscreen_dirty1_vesa_4x_3x_8bpp(void)   { DIRTY1(8)  DIRTY1_NX3 (4,8)  }
void blitscreen_dirty1_vesa_4x_3xs_8bpp(void)  { DIRTY1(8)  DIRTY1_NXNS(4,3,8)  }

void blitscreen_dirty1_vesa_1x_1x_16bpp(void)  { DIRTY1(16) DIRTY1_NXNS(1,1,16) }
void blitscreen_dirty1_vesa_1x_2x_16bpp(void)  { DIRTY1(16) DIRTY1_NX2 (1,16) }
void blitscreen_dirty1_vesa_1x_2xs_16bpp(void) { DIRTY1(16) DIRTY1_NXNS(1,2,16) }
void blitscreen_dirty1_vesa_2x_2x_16bpp(void)  { DIRTY1(16) DIRTY1_NX2 (2,16) }
void blitscreen_dirty1_vesa_2x_2xs_16bpp(void) { DIRTY1(16) DIRTY1_NXNS(2,2,16) }

void blitscreen_dirty0_vesa_1x_1x_8bpp(void)   { DIRTY0(8)  DIRTY0_NXNS(1,1,8)  }
void blitscreen_dirty0_vesa_1x_2x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX2 (1,8)  }
void blitscreen_dirty0_vesa_1x_2xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(1,2,8)  }
void blitscreen_dirty0_vesa_2x_1x_8bpp(void)   { DIRTY0(8)  DIRTY0_NXNS(2,1,8)  }
void blitscreen_dirty0_vesa_2x_2x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX2 (2,8)  }
void blitscreen_dirty0_vesa_2x_2xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(2,2,8)  }
void blitscreen_dirty0_vesa_2x_3x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX3 (2,8)  }
void blitscreen_dirty0_vesa_2x_3xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(2,3,8)  }
void blitscreen_dirty0_vesa_3x_2x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX2 (3,8)  }
void blitscreen_dirty0_vesa_3x_2xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(3,2,8)	}
void blitscreen_dirty0_vesa_3x_3x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX3 (3,8)  }
void blitscreen_dirty0_vesa_3x_3xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(3,3,8)  }
void blitscreen_dirty0_vesa_4x_2x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX2 (4,8)  }
void blitscreen_dirty0_vesa_4x_2xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(4,2,8)	}
void blitscreen_dirty0_vesa_4x_3x_8bpp(void)   { DIRTY0(8)  DIRTY0_NX3 (4,8)  }
void blitscreen_dirty0_vesa_4x_3xs_8bpp(void)  { DIRTY0(8)  DIRTY0_NXNS(4,3,8)  }

void blitscreen_dirty0_vesa_1x_1x_16bpp(void)  { DIRTY0(16) DIRTY0_NXNS(1,1,16) }
void blitscreen_dirty0_vesa_1x_2x_16bpp(void)  { DIRTY0(16) DIRTY0_NX2 (1,16) }
void blitscreen_dirty0_vesa_1x_2xs_16bpp(void) { DIRTY0(16) DIRTY0_NXNS(1,2,16) }
void blitscreen_dirty0_vesa_2x_2x_16bpp(void)  { DIRTY0(16) DIRTY0_NX2 (2,16) }
void blitscreen_dirty0_vesa_2x_2xs_16bpp(void) { DIRTY0(16) DIRTY0_NXNS(2,2,16) }

// Start Emu Plus

void blitscreen_dirty0_vesa_2x_2xs_16bpp_eagle(void)
#define NORM_PIX(LNE,PS1,PIXLNO) \
  newline##LNE##[PS1] = pixel##PIXLNO##l; \
  newline##LNE##[PS1+1] = pixel##PIXLNO##r;

#define BUILDUP(PIXELNO) \
  pixel##PIXELNO## = (pixel##PIXELNO##r << 8) + pixel##PIXELNO##l;

#define GETLR(PIXELNO,STRTPOS, LINENO) \
  pixel##PIXELNO##l = src##LINENO##[STRTPOS]; \
  pixel##PIXELNO##r = src##LINENO##[STRTPOS+1];

#define P12 \
  pixel1 == pixel2
#define P34 \
  pixel3 == pixel4
#define P13 \
  pixel1 == pixel3
#define P24 \
  pixel2 == pixel4

{
	unsigned int pixel1l, pixel1r, pixel2l, pixel2r;
	unsigned int pixel3l, pixel3r, pixel4l, pixel4r;
	unsigned int pixel1, pixel2, pixel3, pixel4;
	unsigned char *src1 = (scrbitmap->line[skiplines] ) + skipcolumns *2 ;
	unsigned char *src2 = (scrbitmap->line[skiplines+1] ) +skipcolumns *2;
	unsigned char newline1[1600], newline2[1600];
	unsigned int y,acnt, lc=4;
	short src_seg=_my_ds(), dest_seg=screen->seg;
	unsigned long address, address2;
	int vesa_line = gfx_yoffset;
	int xoffs = 2*gfx_xoffset;
	int line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);

	address= bmp_write_line(screen,vesa_line)+xoffs;
	copyline_2x_16bpp(src1,dest_seg,address,152);
	src_seg = _my_ds();
	vesa_line++;
	for (y = 0 ; y < 224 ; y++)
	{
		GETLR(1,0,1) ; BUILDUP(1)   // These Macro Things should make it easier to read
		GETLR(2,2,1) ; BUILDUP(2)   // Also eliminate simple problems from typos :)
		GETLR(3,0,2) ; BUILDUP(3)
		GETLR(4,2,2) ; BUILDUP(4)

		NORM_PIX(1,0,1)   // LINE, POSITION OF #1, PIXEL#
		NORM_PIX(2,0,3)

		if ( (P24) && (P34) )
		  { NORM_PIX(1,2,2) } // ** NOTE FOR NEXT ROUTINE, CHANGED TYPE **
		else
		  { NORM_PIX(1,2,1) } // ** NORMAL TYPE **

		if ( (P12) && (P24) )
		  { NORM_PIX(2,2,4) } // ** CHANGED **
		else
		  { NORM_PIX(2,2,3) } // ** NORMAL **

		//  OK, the 'main' loop

		for (acnt = 4 ; acnt < 1216 ; acnt+=4)
		{
			if ( (P13) && (P34) )
			  { NORM_PIX(1,acnt,1) } // ** CHANGE **
			else
			  { NORM_PIX(1,acnt,2) } // ** NORM **

			if ( (P12) && (P13) )
			  { NORM_PIX(2,acnt,3) }  // CHANGE
			else
			  { NORM_PIX(2,acnt,4) }  // NORM

			// Shift

			pixel1 = pixel2 ; pixel1l = pixel2l ; pixel1r = pixel2r;
			pixel3 = pixel4 ; pixel3l = pixel4l ; pixel3r = pixel4r;
			GETLR(2,lc,1) ; BUILDUP(2); // Pixel, Start, Line
			GETLR(4,lc,2) ; BUILDUP(4);
			lc+=2;

			if ( (P24) && (P34) )
			  { NORM_PIX(1,acnt+2,2) }  // CHANGE
			else
			  { NORM_PIX(1,acnt+2,1) }  // NORM

			if ( (P12) && (P24) )
			  { NORM_PIX(2,acnt+2,4)  }  // CHANGE
			else
			  { NORM_PIX(2,acnt+2,3) }  // NORM
		}
		lc = 4;

		if ( (P13) && (P34) )
		  { NORM_PIX(1,acnt+4,1) } // ** CHANGE **
		else
		  { NORM_PIX(1,acnt+4,2) } // ** NORM **

		if ( (P12) && (P13) )
		  { NORM_PIX(2,acnt+4,3) }  // CHANGE
		else
		  { NORM_PIX(2,acnt+4,4) }  // NORM

		NORM_PIX(1,acnt+6,2)   // LINE, POSITION OF #1, PIXEL#
		NORM_PIX(2,acnt+6,4)


		address = bmp_write_line(screen, vesa_line)+xoffs;
		address2 = bmp_write_line(screen,vesa_line+1)+xoffs;

		_movedatal(src_seg, (unsigned long)newline1, dest_seg, address, 304);
		_movedatal(src_seg, (unsigned long)newline2, dest_seg, address2, 304);
		vesa_line +=2;
		src1 = src2;
		src2 += line_offs;
	}
}

void blitscreen_dirty0_vesa_2x_2xs_8bpp_eagle(void)
{
unsigned int vesa_line,line_offs,xoffs;
short src_seg,dest_seg;
char truth1, truth2 ; // Truths, well, they're not Lies :)
unsigned char pixel1, pixel2, pixel3, pixel4 ; // For storing Pixels :/
unsigned int lc;
unsigned int poss=0;
unsigned int y;
unsigned int width = gfx_display_columns;
unsigned char *src = (scrbitmap->line[skiplines]) + skipcolumns;
unsigned char *src2 = (scrbitmap->line[skiplines+1]) +skipcolumns;
unsigned char newline1[1600], newline2[1600]; // Can't have variable sized arrays, set to max it's ever gonna be!, Probably should write direct 2 screen anyway!
unsigned long address, address2;
line_offs = (scrbitmap->line[1]  - scrbitmap->line[0]);
xoffs = gfx_xoffset;
dest_seg = screen->seg;
vesa_line = gfx_yoffset;

// top line

address= bmp_write_line(screen,vesa_line)+xoffs;
copyline_2x_8bpp(src,dest_seg,address,width/4);
src_seg = _my_ds();
vesa_line++;


for (y = 0 ; y < gfx_display_lines ; y++ )
  {
  lc = 0;

  pixel1 = src[lc];
  pixel2 = src[lc+1];
  pixel3 = src2[lc];
  pixel4 = src2[lc+1];

  truth1 = (pixel1 == pixel3);
  truth2 = (pixel2 == pixel4);

  newline1[lc] = pixel1;
  newline2[lc] = pixel3;
  lc++;
  if (truth1)
    {
    newline1[lc] = pixel1;
    newline2[lc] = pixel3;
    }
  else
  if (truth2)
    {
    if (pixel3 == pixel4)
      {
      newline1[lc] = pixel2;
      newline2[lc] = pixel3;
      }
    else
    if (pixel1 == pixel2)
      {
      newline1[lc] = pixel1;
      newline2[lc] = pixel4;
      }
    else
      {
      newline1[lc] = pixel1;
      newline2[lc] = pixel3;
      }
    }
  else
    {
    newline1[lc] = pixel1;
    newline2[lc] = pixel3;
    }

  // That Should Get the First one out the way where there's nowt on the Left
  // Now the 'main' loop

  for (lc = lc; lc<width ; lc++) // Yeah Right ;)
    {
    poss = (2*lc);
    if (truth2)
      {
      newline1[poss] = pixel2;
      newline2[poss] = pixel4;
      }
    else
    if (truth1)
      {
      if (pixel3 == pixel4)
        {
        newline1[poss] = pixel1;
        newline2[poss] = pixel4;
        }
      else
      if (pixel1 == pixel2)
        {
        newline1[poss] = pixel2;
        newline2[poss] = pixel3;
        }
      else
        {
        newline1[poss] = pixel2;
        newline2[poss] = pixel4;
        }
      }
    else
      {
      newline1[poss] = pixel2;
      newline2[poss] = pixel4;
      }

    //  That's the first Part of the Loop, now we need more pixels.

    pixel1 = pixel2;
    pixel3 = pixel4;
    pixel2 = src[lc+1];  // unsure
    pixel4 = src2[lc+1];
    truth1 = truth2 ;
    truth2 = (pixel2 == pixel4);
    poss++;
    //  Now move onto the second bit of the loop

    if (truth1)
      {
      newline1[poss] = pixel1;
      newline2[poss] = pixel3;
      }
    else
    if (truth2)
      {
      if (pixel3 == pixel4)
        {
        newline1[poss] = pixel2;
        newline2[poss] = pixel3;
        }
      else
      if (pixel1 == pixel2)
        {
        newline1[poss] = pixel1;
        newline2[poss] = pixel4;
        }
      else
        {
        newline1[poss] = pixel1;
        newline2[poss] = pixel3;
        }
      }
    else
      {
      newline1[poss] = pixel1;
      newline2[poss] = pixel3;
      }

    // Ok end of 'main' bit, now to deal with the last pixel.
    }  // End of inner loop marker
  //  Last Pixel Start
   poss++;
  if (pixel2 == pixel4)  // a truth?
    {
    newline1[poss] = pixel2;
    newline2[poss] = pixel4;
    }
  else
  if (pixel1 == pixel3) // truth again?
    {
    if (pixel3 == pixel4)
      {
      newline1[poss] = pixel1;
      newline2[poss] = pixel4;
      }
    else
    if (pixel1 == pixel2)
      {
      newline1[poss] = pixel2;
      newline2[poss] = pixel3;
      }
    else
      {
      newline1[poss] = pixel2;
      newline2[poss] = pixel4;
      }
    }
  else
    {
    newline1[poss] = pixel2;
    newline2[poss] = pixel4;
    }
  //  Yippee Should now have 2 lines :)
  address= bmp_write_line(screen,vesa_line)+xoffs;
  address2 = bmp_write_line(screen, vesa_line+1)+xoffs;
 // src_seg = _my_ds();
  _movedatal(src_seg,(unsigned long)newline1, dest_seg, address, width/2);
  _movedatal(src_seg,(unsigned long)newline2, dest_seg, address2, width/2);

  vesa_line += 2;
  src = src2;
  src2 += line_offs;

  } // End of vertical loop marker

}  // Mm end of function :)  probably 20 decades after it was called ;)

void blitscreen_dirty0_vesa_1x_1x_16bpp_craptv(void)
{
  short dest_seg;
  int y,vesa_line,line_offs,xoffs,width;
  unsigned char *src;
  unsigned long address;

  int dot, dot2;  // DH!
  char dst [2];   // Temp!
  int i;
  unsigned int ac ;

  dest_seg = screen->seg;
  vesa_line = gfx_yoffset;
  line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);
  xoffs = 2*(gfx_xoffset + triplebuf_pos);
  width = gfx_display_columns/2;
  src = scrbitmap->line[skiplines] + 2*skipcolumns;

  for (y = 0;y < gfx_display_lines;y++)
  {
  // Been a long time!  Hmmm not too sure about all this tripplebuffer :|
  // CrappyTv from Daves (dtmnt) Dgen -:Decent Genesis Emu:-

  address = bmp_write_line(screen,vesa_line) + xoffs;
  _farsetsel (dest_seg);

  for (ac = 0 ; ac < gfx_display_columns*2-2 ; ac+=2)
  {

  dot  = src[ ac     ] | ( src[ ac + 1 ] << 8 ) ;
  dot2 = src[ ac + 2 ] | ( src[ ac + 3 ] << 8 ) ;

  dot  = ( ( dot  & 0x07E0 ) << 16 ) | ( dot  & 0xF81F ) ;
  dot2 = ( ( dot2 & 0x07E0 ) << 16 ) | ( dot2 & 0xF81F ) ;

  dot  = ( dot + dot2 ) >> 1 ;

  dot  =  ( ( dot >> 16 ) & 0x07E0 ) | ( dot  & 0xF81F ) ;

  //dst[ 0 ] =   dot        & 255 ;
  //dst[ 1 ] = ( dot >> 8 ) & 255 ;
  _farnspokel (address, dot);
  //_farnspokel (address+1 , dst[1]);
  address+=2;
  }
  vesa_line += 1;
  src += line_offs;
  }

  if (use_triplebuf)
  {
  vesa_scroll_async(triplebuf_pos,0);
  triplebuf_pos = (triplebuf_pos + triplebuf_page_width) % (3*triplebuf_page_width);
  }

}

void blitscreen_dirty0_vesa_2x_2x_16bpp_craptv(void)
{
  short dest_seg;
  int y,vesa_line,line_offs,xoffs,width;
  unsigned char *src;
  unsigned long address;

  int dot, dot2;  // DH!
  char dst [2];   // Temp!
  int i;
  unsigned int ac ;


/*		address = bmp_write_line(screen,vesa_line) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \
		address = bmp_write_line(screen,vesa_line+1) + xoffs; \
		copyline_##MX##x_##BPP##bpp(src,dest_seg,address,width); \*/

  dest_seg = screen->seg;
  vesa_line = gfx_yoffset;
  line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);
  xoffs = 2*(gfx_xoffset + triplebuf_pos);
  width = gfx_display_columns/2;
  src = scrbitmap->line[skiplines] + 2*skipcolumns;

  for (y = 0;y < gfx_display_lines;y++)
  {
  // Been a long time!  Hmmm not too sure about all this tripplebuffer :|
  // CrappyTv from Daves (dtmnt) Dgen -:Decent Genesis Emu:-

  address = bmp_write_line(screen,vesa_line) + xoffs;
  _farsetsel (dest_seg);

  for (ac = 0 ; ac < gfx_display_columns*2-2 ; ac+=2)
  {

  dot  = src[ ac     ] | ( src[ ac + 1 ] << 8 ) ;
  dot2 = src[ ac + 2 ] | ( src[ ac + 3 ] << 8 ) ;

  dot  = ( ( dot  & 0x07E0 ) << 16 ) | ( dot  & 0xF81F ) ;
  dot2 = ( ( dot2 & 0x07E0 ) << 16 ) | ( dot2 & 0xF81F ) ;

  dot  = ( dot + dot2 ) >> 1 ;

  dot  =  ( ( dot >> 16 ) & 0x07E0 ) | ( dot  & 0xF81F ) ;

  //dst[ 0 ] =   dot        & 255 ;
  //dst[ 1 ] = ( dot >> 8 ) & 255 ;
  _farnspokel (address, dot);
  //_farnspokel (address+1 , dst[1]);
  address+=4;
  }
  vesa_line += 2;
  src += line_offs;
  }

  if (use_triplebuf)
  {
  vesa_scroll_async(triplebuf_pos,0);
  triplebuf_pos = (triplebuf_pos + triplebuf_page_width) % (3*triplebuf_page_width);
  }

}

void blitscreen_dirty0_vesa_1x_1x_16bpp_lcdblur(void)  {
	short dest_seg;
	int y,vesa_line,line_offs,xoffs,width;
	unsigned char *src, *src2;
	unsigned long address;
	unsigned int xcount;
	unsigned int topleft, topright, botleft, botright;

	unsigned char avred, avgrn, avblu;
		unsigned char newavred, newavgrn, newavblu;
	unsigned int blurcount;
	unsigned int pixelclr;
	extern int cleanblur;

	dest_seg = screen->seg;
	vesa_line = gfx_yoffset;
	line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);
	xoffs = (16/8)*(gfx_xoffset + triplebuf_pos);
	width = gfx_display_columns/(4/(16/8));
	src = scrbitmap->line[skiplines] + (16/8)*skipcolumns;
	src2 = scrbitmap->line[skiplines+1] + (16/8)*skipcolumns;
	
    blurcount = 0;
	for (y = 0;y < (gfx_display_lines/2)+1;y++) 
	{
		address = bmp_write_line (screen,vesa_line  ) + xoffs;
	
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=4)
		{

		topleft  = (src[xcount]) + (src[xcount+1] << 8);
		topright = (src[xcount+2]) + (src[xcount+3] << 8);
		botleft  = (src2[xcount]) + (src2[xcount+1] << 8);
		botright = (src2[xcount+2]) + (src2[xcount+3] << 8);

       avred = ((((topleft & 0xf800) >> 11)) + (((topright & 0xf800) >> 11)) + (((botleft & 0xf800) >> 11)) + (((botright & 0xf800) >> 11))) >> 2;
       avgrn = ((((topleft & 0x07e0) >> 5 )) + (((topright & 0x07e0) >> 5 )) + (((botleft & 0x07e0) >> 5 )) + (((botright & 0x07e0) >> 5 ))) >> 2;
       avblu = ((((topleft & 0x001f) >> 0 )) + (((topright & 0x001f) >> 0 )) + (((botleft & 0x001f) >> 0 )) + (((botright & 0x001f) >> 0 ))) >> 2;

	   newavred = ( avred + 3*redpoin[blurcount] ) >> 2 ;
       newavgrn = ( avgrn + 3*grnpoin[blurcount] ) >> 2 ;
       newavblu = ( avblu + 3*blupoin[blurcount] ) >> 2 ;
	   if (cleanblur) {
	   if (newavred == redpoin [ blurcount ]) newavred=avred /*avred*/; 
       if (newavgrn == grnpoin [ blurcount ]) newavgrn=avgrn /*=avgrn*/;
       if (newavblu == blupoin [ blurcount ]) newavblu=avblu /*=avblu*/;
       }
       redpoin [ blurcount ] = newavred;
       grnpoin [ blurcount ] = newavgrn;
       blupoin [ blurcount ] = newavblu;
	   pixelclr = ( newavblu << 0 ) + (newavgrn << 5) + (newavred << 11);
	
  		_farnspokel (address+(xcount/2), pixelclr);
	
         blurcount++;

		}
		vesa_line ++;
		src += line_offs*2;
		src2+= line_offs*2;
	}
	TRIPLEBUF_FLIP
}

void blitscreen_dirty0_vesa_2x_2x_16bpp_lcdblur(void)  {
	short dest_seg;
	int y,vesa_line,line_offs,xoffs,width;
	unsigned char *src, *src2;
	unsigned long address;
	unsigned int xcount;
	unsigned int topleft, topright, botleft, botright;

	unsigned char avred, avgrn, avblu;
		unsigned char newavred, newavgrn, newavblu;
	unsigned int blurcount;
	unsigned int pixelclr;
	extern int cleanblur;

	dest_seg = screen->seg;
	vesa_line = gfx_yoffset;
	line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);
	xoffs = (16/8)*(gfx_xoffset + triplebuf_pos);
	width = gfx_display_columns/(4/(16/8));
	src = scrbitmap->line[skiplines] + (16/8)*skipcolumns;
	src2 = scrbitmap->line[skiplines+1] + (16/8)*skipcolumns;
	
    blurcount = 0;
	for (y = 0;y < (gfx_display_lines/2)+1;y++) 
	{
		address = bmp_write_line (screen,vesa_line  ) + xoffs;
					
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=4)
		{

		topleft  = (src[xcount]) + (src[xcount+1] << 8);
		topright = (src[xcount+2]) + (src[xcount+3] << 8);
		botleft  = (src2[xcount]) + (src2[xcount+1] << 8);
		botright = (src2[xcount+2]) + (src2[xcount+3] << 8);

       avred = ((((topleft & 0xf800) >> 11)) + (((topright & 0xf800) >> 11)) + (((botleft & 0xf800) >> 11)) + (((botright & 0xf800) >> 11))) >> 2;
       avgrn = ((((topleft & 0x07e0) >> 5 )) + (((topright & 0x07e0) >> 5 )) + (((botleft & 0x07e0) >> 5 )) + (((botright & 0x07e0) >> 5 ))) >> 2;
       avblu = ((((topleft & 0x001f) >> 0 )) + (((topright & 0x001f) >> 0 )) + (((botleft & 0x001f) >> 0 )) + (((botright & 0x001f) >> 0 ))) >> 2;

	   newavred = ( avred + 3*redpoin[blurcount] ) >> 2 ;
       newavgrn = ( avgrn + 3*grnpoin[blurcount] ) >> 2 ;
       newavblu = ( avblu + 3*blupoin[blurcount] ) >> 2 ;
	   if (cleanblur) {
	   if (newavred == redpoin [ blurcount ]) newavred=avred /*avred*/; 
       if (newavgrn == grnpoin [ blurcount ]) newavgrn=avgrn /*=avgrn*/;
       if (newavblu == blupoin [ blurcount ]) newavblu=avblu /*=avblu*/;
       }
       redpoin [ blurcount ] = newavred;
       grnpoin [ blurcount ] = newavgrn;
       blupoin [ blurcount ] = newavblu;
	   pixelclr = ( newavblu << 0 ) + (newavgrn << 5) + (newavred << 11);
	
  		_farnspokel (address+(xcount*2), pixelclr);
		_farnspokel (address+(xcount*2)+4, pixelclr);
		_farnspokel (address+(xcount*2)+8, pixelclr);
		_farnspokel (address+(xcount*2)+12, pixelclr);
         blurcount++;

		}
		address = bmp_write_line (screen,vesa_line+1) + xoffs;
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=4)
		{

		topleft  = (src[xcount]) + (src[xcount+1] << 8);
		topright = (src[xcount+2]) + (src[xcount+3] << 8);
		botleft  = (src2[xcount]) + (src2[xcount+1] << 8);
		botright = (src2[xcount+2]) + (src2[xcount+3] << 8);

       avred = ((((topleft & 0xf800) >> 11)) + (((topright & 0xf800) >> 11)) + (((botleft & 0xf800) >> 11)) + (((botright & 0xf800) >> 11))) >> 2;
       avgrn = ((((topleft & 0x07e0) >> 5 )) + (((topright & 0x07e0) >> 5 )) + (((botleft & 0x07e0) >> 5 )) + (((botright & 0x07e0) >> 5 ))) >> 2;
       avblu = ((((topleft & 0x001f) >> 0 )) + (((topright & 0x001f) >> 0 )) + (((botleft & 0x001f) >> 0 )) + (((botright & 0x001f) >> 0 ))) >> 2;

	   newavred = ( avred + 3*redpoin[blurcount] ) >> 2 ;
       newavgrn = ( avgrn + 3*grnpoin[blurcount] ) >> 2 ;
       newavblu = ( avblu + 3*blupoin[blurcount] ) >> 2 ;
	   if (cleanblur) {
	   if (newavred == redpoin [ blurcount ]) newavred=avred /*avred*/; 
       if (newavgrn == grnpoin [ blurcount ]) newavgrn=avgrn /*=avgrn*/;
       if (newavblu == blupoin [ blurcount ]) newavblu=avblu /*=avblu*/;
       }
       redpoin [ blurcount ] = newavred;
       grnpoin [ blurcount ] = newavgrn;
       blupoin [ blurcount ] = newavblu;
	   pixelclr = ( newavblu << 0 ) + (newavgrn << 5) + (newavred << 11);
	
  		_farnspokel (address+(xcount*2), pixelclr);
		_farnspokel (address+(xcount*2)+4, pixelclr);
		_farnspokel (address+(xcount*2)+8, pixelclr);
		_farnspokel (address+(xcount*2)+12, pixelclr);
         blurcount++;
		}
		address = bmp_write_line (screen,vesa_line+2) + xoffs;
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=4)
		{

		topleft  = (src[xcount]) + (src[xcount+1] << 8);
		topright = (src[xcount+2]) + (src[xcount+3] << 8);
		botleft  = (src2[xcount]) + (src2[xcount+1] << 8);
		botright = (src2[xcount+2]) + (src2[xcount+3] << 8);

       avred = ((((topleft & 0xf800) >> 11)) + (((topright & 0xf800) >> 11)) + (((botleft & 0xf800) >> 11)) + (((botright & 0xf800) >> 11))) >> 2;
       avgrn = ((((topleft & 0x07e0) >> 5 )) + (((topright & 0x07e0) >> 5 )) + (((botleft & 0x07e0) >> 5 )) + (((botright & 0x07e0) >> 5 ))) >> 2;
       avblu = ((((topleft & 0x001f) >> 0 )) + (((topright & 0x001f) >> 0 )) + (((botleft & 0x001f) >> 0 )) + (((botright & 0x001f) >> 0 ))) >> 2;

	   newavred = ( avred + 3*redpoin[blurcount] ) >> 2 ;
       newavgrn = ( avgrn + 3*grnpoin[blurcount] ) >> 2 ;
       newavblu = ( avblu + 3*blupoin[blurcount] ) >> 2 ;
	   if (cleanblur) {
	   if (newavred == redpoin [ blurcount ]) newavred=avred /*avred*/; 
       if (newavgrn == grnpoin [ blurcount ]) newavgrn=avgrn /*=avgrn*/;
       if (newavblu == blupoin [ blurcount ]) newavblu=avblu /*=avblu*/;
       }
       redpoin [ blurcount ] = newavred;
       grnpoin [ blurcount ] = newavgrn;
       blupoin [ blurcount ] = newavblu;
	   pixelclr = ( newavblu << 0 ) + (newavgrn << 5) + (newavred << 11);
	
  		_farnspokel (address+(xcount*2), pixelclr);
		_farnspokel (address+(xcount*2)+4, pixelclr);
		_farnspokel (address+(xcount*2)+8, pixelclr);
		_farnspokel (address+(xcount*2)+12, pixelclr);
         blurcount++;
		}
		address = bmp_write_line (screen,vesa_line+3) + xoffs;
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=4)
		{

		topleft  = (src[xcount]) + (src[xcount+1] << 8);
		topright = (src[xcount+2]) + (src[xcount+3] << 8);
		botleft  = (src2[xcount]) + (src2[xcount+1] << 8);
		botright = (src2[xcount+2]) + (src2[xcount+3] << 8);

       avred = ((((topleft & 0xf800) >> 11)) + (((topright & 0xf800) >> 11)) + (((botleft & 0xf800) >> 11)) + (((botright & 0xf800) >> 11))) >> 2;
       avgrn = ((((topleft & 0x07e0) >> 5 )) + (((topright & 0x07e0) >> 5 )) + (((botleft & 0x07e0) >> 5 )) + (((botright & 0x07e0) >> 5 ))) >> 2;
       avblu = ((((topleft & 0x001f) >> 0 )) + (((topright & 0x001f) >> 0 )) + (((botleft & 0x001f) >> 0 )) + (((botright & 0x001f) >> 0 ))) >> 2;

	   newavred = ( avred + 3*redpoin[blurcount] ) >> 2 ;
       newavgrn = ( avgrn + 3*grnpoin[blurcount] ) >> 2 ;
       newavblu = ( avblu + 3*blupoin[blurcount] ) >> 2 ;
	   if (cleanblur) {
	   if (newavred == redpoin [ blurcount ]) newavred=avred /*avred*/; 
       if (newavgrn == grnpoin [ blurcount ]) newavgrn=avgrn /*=avgrn*/;
       if (newavblu == blupoin [ blurcount ]) newavblu=avblu /*=avblu*/;
       }
       redpoin [ blurcount ] = newavred;
       grnpoin [ blurcount ] = newavgrn;
       blupoin [ blurcount ] = newavblu;
	   pixelclr = ( newavblu << 0 ) + (newavgrn << 5) + (newavred << 11);
	
  		_farnspokel (address+(xcount*2), pixelclr);
		_farnspokel (address+(xcount*2)+4, pixelclr);
		_farnspokel (address+(xcount*2)+8, pixelclr);
		_farnspokel (address+(xcount*2)+12, pixelclr);

	
         blurcount++;
		}
		vesa_line += 4;
		src += line_offs*2;
		src2+= line_offs*2;
	}
	TRIPLEBUF_FLIP
}

void blitscreen_dirty0_vesa_2x_2x_16bpp_50scan(void) {
	
	short dest_seg;
	int y,vesa_line,line_offs,xoffs,width;
	unsigned char *src;
	unsigned long address, address2;
	unsigned int xcount;
	unsigned char valblu, valgrn, valred;
    unsigned long norm, darkened;
	dest_seg = screen->seg;
	vesa_line = gfx_yoffset;
	line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);
	xoffs = (16/8)*(gfx_xoffset + triplebuf_pos);
	width = gfx_display_columns/(4/(16/8));
	src = scrbitmap->line[skiplines] + (16/8)*skipcolumns;
	for (y = 0;y < gfx_display_lines;y++) 
	{
		address = bmp_write_line (screen,vesa_line  ) + xoffs;
		address2 = bmp_write_line(screen,vesa_line+1) + xoffs;
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=2)
		{

		norm = (src[xcount]) + (src[xcount+1] << 8);
	    norm = norm + (norm << 16);
	
        valblu = ((norm & 0x001f) >> 0 ) >> 1;
        valgrn = ((norm & 0x07e0) >> 5 ) >> 1;
        valred = ((norm & 0xf800) >> 11) >> 1;
		darkened = ( valblu << 0 ) + (valgrn << 5) + (valred << 11);
		darkened = darkened + (darkened << 16);
		_farnspokel (address+(xcount*2), norm);     // Slow!
		_farnspokel (address2+(xcount*2),darkened); // Slow!
		}
		vesa_line += 2;
		src += line_offs;
	}
	TRIPLEBUF_FLIP
}

void blitscreen_dirty0_vesa_1x_1x_16bpp_50scan(void) {
	
	short dest_seg;
	int y,vesa_line,line_offs,xoffs,width;
	unsigned char *src;
	unsigned long address, address2;
	unsigned int xcount;
	unsigned char valblu, valgrn, valred;
    unsigned long norm, darkened;
	dest_seg = screen->seg;
	vesa_line = gfx_yoffset;
	line_offs = (scrbitmap->line[1] - scrbitmap->line[0]);
	xoffs = (16/8)*(gfx_xoffset + triplebuf_pos);
	width = gfx_display_columns/(4/(16/8));
	src = scrbitmap->line[skiplines] + (16/8)*skipcolumns;
	for (y = 0;y < gfx_display_lines;y++) 
	{
		address = bmp_write_line (screen,vesa_line  ) + xoffs;
		address2 = bmp_write_line(screen,vesa_line+1) + xoffs;
		_farsetsel(dest_seg);
		for (xcount = 0; xcount < width*4 ; xcount+=2)
		{

		norm = (src[xcount]) + (src[xcount+1] << 8);
	    norm = norm + (norm << 16);
	
        valblu = ((norm & 0x001f) >> 0 ) >> 1;
        valgrn = ((norm & 0x07e0) >> 5 ) >> 1;
        valred = ((norm & 0xf800) >> 11) >> 1;
		darkened = ( valblu << 0 ) + (valgrn << 5) + (valred << 11);
		darkened = darkened + (darkened << 16);
		_farnspokel (address+(xcount), norm);     // Slow!
		_farnspokel (address2+(xcount),darkened); // Slow!
		}
		vesa_line += 1;
		src += line_offs;
	}
	TRIPLEBUF_FLIP
}
// End Emu Plus



